/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

//Base Small Placer
// - Small Placement Problem
//   created by Andrew Caldwell  981206

#include "branchPl.h"

BranchSmallPlacer::BranchSmallPlacer(SmallPlacementProblem& problem)
	: BaseSmallPlacer(problem), 
	  _c(problem.getNumCells()+2),
	  _curSoln(problem.getNumCells()), 
	  _bestSolnCostStack(problem.getNumCells()+1, DBL_MAX)
{
    abkfatal(problem.getNumRows() == 1, 
	"BB SmallPlacer can't deal with multiple rows");
    abkfatal(_problem.getSolnBuffer()[0].populated,
	"BB SmallPlacer doesn't know how to use an initial placement");

    netState compilerBugInBRSmallPlacer(XInterval(0,0), 0); 

	//set all orientations to N
    vector<Orient>& ori = _problem.getSolnBuffer()[0].orientations;
    fill(ori.begin(), ori.end(), Orientation());

    Placement& soln     = _problem.getSolnBuffer()[0].placement;
    const unsigned numCells = _problem.getNumCells();
    const unsigned numNets  = _problem.getNumNets();


	//setup the stack and queue to produce the ordering
    _s.reserve(numCells);

    _c[numCells]   = numCells;
    _c[numCells+1] = numCells+1;
    _curWL = 0;
 
    for(_idx = numCells; _idx != 0; )
   	_q.push_back(--_idx); 

    const XInterval rowIvl(problem.getRows()[0].xMin,problem.getRows()[0].xMax);

    for(_idx = numNets; _idx != 0; )
    {
	_idx--;
	
	_netStates[_idx].reserve(numCells);

	_netStates[_idx].push_back(
	 netState(XInterval(problem.getNetTerminalBBox(_idx).xMin,
			    problem.getNetTerminalBBox(_idx).xMax), 
			    _minNetCost));

	XInterval& netIvl = _netStates[_idx].back().first;

	if(!netIvl.isEmpty())
	{
	   if(netIvl.max < rowIvl.min)
		netIvl.max = rowIvl.min; //same as += rowInv.min
	   else if(netIvl.min > rowIvl.max)
		netIvl.min = rowIvl.max; //same as += rowInv.max

	   _netStates[_idx].back().second = 
			max(_minNetCost, netIvl.getLength());
	}

	_curWL += _netStates[_idx].back().second;
    }

    _idx = numCells-1;


    _nxtLoc = Point(problem.getRows()[0].xMin, problem.getRows()[0].yMin);
    double bestWL = DBL_MAX;

    const unsigned boundLookAhead = 2;
    bool  bestIsSet = false;

    while(_idx < numCells)
    {
    	addCell();

	if(bestIsSet &&
	  ((_curWL > _bestSolnCostStack[_costStack.size()+boundLookAhead]) || 
	   (_curWL >= bestWL)))
	{
	   /* 
	    cout<<" idx: "<<_idx<<endl;
	    cout<<" curWL: "<<_curWL<<"   prev best was "<<
		_bestSolnCostStack[_costStack.size()]<<endl;
	    cout<<" costStack: "<<_costStack;
	    cout<<" Seq: "<<_s<<endl;
	   */ 
	    _c[_idx] = 0;
	}

        if(_c[_idx] == 0)
      	{
	    if(_curWL < bestWL && _idx == 0)
	    { 
		bestIsSet = true;
		bestWL = _curWL;

		for(unsigned cell = numCells; cell != 0; )
		{ cell--; soln[cell] = _curSoln[cell];}
    	        _problem.getSolnBuffer()[0].populated = true;

		_bestSolnCostStack = _costStack;

		_bestSolnCostStack.push_back(bestWL);
		_bestSolnCostStack.push_back(bestWL);
		for(int b = boundLookAhead-1; b >= 0; b--)
		    _bestSolnCostStack.push_back(bestWL);
		/*
		cout<<"Best Cost Stack is "<<endl<<_bestSolnCostStack;
		cout<<"CurWL "<<_curWL<<endl;
		cout<<"Sequence "<<_s<<endl;
		*/
	    }

            while(_c[_idx] == 0)
            {
	   	if(_idx < numCells)
		  removeCell();

           	_c[++_idx]--; 
            }
      	}
      	_idx--;
    }

    cout<<"Done with B&B"<<endl;
}

void BranchSmallPlacer::addCell()
{
    _s.push_back(_q.front());
    _q.pop_front();
    _c[_idx] = _idx;
    _costStack.push_back(_curWL);

    const unsigned cell = _s.back();
    _curSoln[cell] = _nxtLoc;

    //push the new values onto the net's stacks. update the running cost
    //as we go.

    const unsigned endIdx = cellConnectionsEnd(cell);
    for(unsigned nIdx = cellConnectionsBegin(cell); nIdx != endIdx; nIdx++)
    {
	unsigned n = _connections[nIdx];

	_curWL -= _netStates[n].back().second; //sub the previous WL 

	_netStates[n].push_back(_netStates[n].back());

	XInterval& newIvl = _netStates[n].back().first;
	newIvl += (_nxtLoc.x + netlist.getPinOffset(cell,n).x);

		//max'ing with minNetCost discourages edges with 
		//0 horizontal cost.
	_netStates[n].back().second = max(_minNetCost, newIvl.getLength());

	_curWL += _netStates[n].back().second; //add the new WL
    }

    _nxtLoc.x += _bloatWidths[cell];
}

void BranchSmallPlacer::removeCell()
{
    unsigned cell = _s.back();

    _q.push_back(cell);
    _s.pop_back();
    _nxtLoc.x -= _bloatWidths[cell];
    _curWL = _costStack.back();
    _costStack.pop_back();

    const unsigned endIdx = cellConnectionsEnd(cell);
    for(unsigned nIdx = cellConnectionsBegin(cell); nIdx != endIdx; nIdx++)
	_netStates[_connections[nIdx]].pop_back();
}

void BranchSmallPlacer::print()
{
    cout<<"Q: ";
    for(unsigned i = 0; i < _q.size(); i++)
	cout<<_q[i]<<" ";
    cout<<endl; 

    cout<<"S: ";
    for(i = 0; i < _s.size(); i++)
	cout<<_s[i]<<" ";
    cout<<endl; 

    cout<<"C: ";
    for(i = _c.size(); i != 0; )
	cout<<_c[--i]<<" ";
    cout<<endl;

    cout<<"Idx: "<<_idx<<endl<<endl;
}


