/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/


//created by Andrew Caldwell on 12/12/99 caldwell@cs.ucla.edu

#include "kPlSplitter.h"
#include "splitSmall.h"
#include "splitLarge.h"
#include "partProbForCapo.h"
#include "partitionIds.h"
#include "netCutWBits.h"

class CompNodesByPlX
{
    const Placement& _pl;
  public:
    CompNodesByPlX(const Placement& pl) : _pl(pl) {}

    bool operator()(unsigned n1, unsigned n2)
    {	return _pl[n1].x < _pl[n2].x;}
};

KPlaceSplitter::KPlaceSplitter(CapoBlock&  		blockToSplit,
                   	       const CapoPlacer&	capo,
			       const Placement&	 	origPl,
                   	       CapoSplitterParams      	params)
        : BaseBlockSplitter(blockToSplit, capo, params),
          _origPl(origPl), _netlistHG(capo.getNetlistHGraph())
{

cout<<" Using k-place based splitter on block: "<<_block<<endl;

    //find the center of mass of the cells in the block

    vector<unsigned>::const_iterator cId;
    Point  centerOfMass(0,0);
    double totalWeight = 0;

    for(cId = _block.cellIdsBegin(); cId != _block.cellIdsEnd(); cId++)
    {
	double  cellWt = _netlistHG.getNodeByIdx(*cId).getWeight();
	centerOfMass.x   += cellWt * _origPl[*cId].x;
	centerOfMass.y   += cellWt * _origPl[*cId].y;
	totalWeight      += cellWt;
    }

    centerOfMass.x /= totalWeight;
    centerOfMass.y /= totalWeight;

    BBox centerBBox = _block.getBBox();
    centerBBox.ShrinkTo(0.5);

    bool plSplitWorked = false;

    cout<<" Block BBox:   "<<_block.getBBox()<<endl;
    cout<<" Block Middle: "<<centerBBox<<endl;
    cout<<" Center:       "<<centerOfMass<<"("
	<<100*(centerOfMass.x-_block.getBBox().xMin)/
		_block.getBBox().getWidth()
	<<" %)"<<endl;

    if(!centerBBox.contains(centerOfMass))
	cout<<"Center is not inside block: "<<endl;

    if(centerBBox.contains(centerOfMass))
    {
	cout<<"Center is inside block middle"<<endl;


	//generate an ordering of the nodes, based on the placement
	vector<unsigned> sortedNodes;
	sortedNodes.insert(sortedNodes.begin(), _block.cellIdsBegin(),
			  _block.cellIdsEnd());
	sort(sortedNodes.begin(), sortedNodes.end(), CompNodesByPlX(_origPl));

	//double   totalWt = _block.getTotalCellArea();
	double   xMiddle = _block.getBBox().xMin + 
			_block.getBBox().getWidth()/2.0;

        PartitioningProblemForCapo problem(_block, capo, false, xMiddle);
        Partitioning& constr = 
			const_cast<Partitioning&>(problem.getFixedConstr());

	unsigned numToFix = sortedNodes.size()/5; //fix left and right 5%
	const SubHGraph& subHG = problem.getSubHGraph();

	for(unsigned n = 0; n < numToFix; n++)
	{
	    unsigned nId     = sortedNodes[n];
	    unsigned subHGId = subHG.origNode2NewIdx(nId);
	    constr[subHGId].setToPart(0);
	}
	for(n = sortedNodes.size()-1; n > sortedNodes.size()-numToFix; n--)
	{
	    unsigned nId     = sortedNodes[n];
	    unsigned subHGId = subHG.origNode2NewIdx(nId);
	    constr[subHGId].setToPart(1);
	}

	problem.setTolerance(10);	//10% tolerance
	MLPart::Parameters mlParams;
 	MLPart partitioner(problem, mlParams);

	vector<double>      partAreas(2);
	partAreas[0] = partitioner.getPartitionArea(0);
        partAreas[1] = partitioner.getPartitionArea(1);

	vector<double> siteAreas(2,0.0);
	double _xSplit     = _block.findXSplit(partAreas[0], partAreas[1],
                                             5, siteAreas);

        cout<<" Best Solution had cut "<<partitioner.getBestSolnCost()<<endl;

    	createVSplitBlocks(problem, _xSplit);
	cout<<" KPlace splitter was successful"<<endl;
	plSplitWorked = true;
    }
    if(!plSplitWorked) //either too clumpy, or CenterMass outside the OK range
    {	//call the VSplitter

	cout<<"  Using the partitioning splitter"<<endl;
	if(_block.getNumCells() < 200)
        {
	    SplitSmallBlockVertically partSplitter(_block, _capo, _params.verb);
	    _newBlocks = partSplitter.getNewBlocks();
	}
	else
	{
	    SplitLargeBlockVertically partSplitter(_block, _capo, _params.verb);
	    _newBlocks = partSplitter.getNewBlocks();
	}
    }
}

