/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/


//created by Andrew Caldwell on 10/17/99 caldwell@cs.ucla.edu


#include "splitSmall.h"
#include "partProbForCapo.h"
#include "mlPart.h"
#include "fmPart.h"
#include "bbFMPart.h"

SplitSmallBlockHorizontally::SplitSmallBlockHorizontally
	(      CapoBlock&  blockToSplit,
	 const CapoPlacer& capo,
	       CapoSplitterParams params)
	: BaseBlockSplitter(blockToSplit,capo,params)
{
    abkfatal(_block.getNumRows() > 1,	
	"can't split a 1 row block horizontally");

    if(_params.verb.forActions > 2)
	cout<<"Splitting a small block horizontally"<<endl;
   
   	//notes:

	//shifting the cutline will be difficult on such small
	//problems, so we won't allow as much tolerance.
	//also, with the small number of rows small problems have,
	//shifting by even 1 row would mean a large change in
	//tolerance

	//P0 is the top partition

    const vector<const RBPCoreRow*>& rows = _block.getRows();
    vector<double>   targetCellAreas(2,_block.getTotalCellArea()/2.0);
    vector<double>   actualCaps(2,0);

    _splitRow  = findBestSplitRow(targetCellAreas, actualCaps);

    if(_params.verb.forActions > 4)
        cout<<"P0/P1 rows: "<<rows.size()-_splitRow<<"/"<<_splitRow<<endl;

    PartitioningProblemForCapo problem(_block, capo, true, 
					rows[_splitRow]->getYCoord());

    if(_params.doRepartitioning || _params.useWSTolMethod)
	problem.setPaperMethodTolerance(_splitRow);
    else
	problem.setTolerance(_params.constantTolerance);

    callPartitioner(problem);

    if(_params.verb.forActions > 8)
	cout<<"Done partitioning...creating blocks"<<endl;

    createHSplitBlocks(problem, _splitRow);
}
 
    
void SplitSmallBlockHorizontally::callPartitioner
		(PartitioningProblemForCapo& problem)
{
    bool foundGoodSplit = false;
    unsigned numTries   = 0;
   
    while(!foundGoodSplit && numTries++ < 2)
    {
	if(_block.getNumCells() <= _capo.getParams().smallPartThreshold)
	{
	    if(_params.verb.forActions > 8)
		cout<<"Using BBPart"<<endl;
	    foundGoodSplit = callBBFM(problem);
	}
	else
	{
	    if(_params.verb.forActions > 8)
		cout<<"Using FMPart"<<endl;
	    foundGoodSplit = callFMPart(problem);

	    if(!foundGoodSplit && numTries < 2)
	    {
	 	if(_params.verb.forActions > 9)
		    cout<<"  Failed. clearing part buffers"<<endl;
		PartitioningBuffer& buff = problem.getSolnBuffers();
		unsigned numTerms=problem.getHGraph().getNumTerminals();
		unsigned sol;
	        for(sol = buff.beginUsedSoln(); 
		    sol != buff.endUsedSoln(); sol++)
		{
		    Partitioning& part = buff[sol];
	 	    for(unsigned i = numTerms; i < part.size(); i++)
			part[i].setToAll(2);
		}

		//consider changing the problem here to attempt to
		//get a better solution	
	    }
	}
    }
}

bool SplitSmallBlockHorizontally::callBBFM
		(PartitioningProblemForCapo& problem)
{
    problem.reserveBuffers(2);

    BBFMPart::Parameters bbfmparams;
    bbfmparams.fmStarts = 4;
    bbfmparams.bbParams.pushLimit = 1000000; //~2 seconds

    problem.getSolnBuffers().setBeginUsedSoln(0);
    problem.getSolnBuffers().setEndUsedSoln(1);

    Timer smallPartTimer;
    BBFMPart partitioner1(problem, bbfmparams);
    smallPartTimer.stop();


    //if even # rows or hard problem, take this soln
    if((_block.getNumRows()%2==0) || smallPartTimer.getUserTime() > 0.1)
	return true;

    unsigned origSplitRow = _splitRow;

	//otherwise, try rounding the split the other way..
    if(_splitRow <= floor(_block.getNumRows()/2.0))
	_splitRow++;
    else
	_splitRow--;

    abkfatal(_splitRow != 0 && _splitRow < _block.getNumRows(),
               "new blk has no rows");

    problem.setPaperMethodTolerance(_splitRow);
    problem.getSolnBuffers().setBeginUsedSoln(1);
    problem.getSolnBuffers().setEndUsedSoln(2);
	     
    BBFMPart partitioner2(problem, bbfmparams);

		//note to self...could include area-balances
	 	//and pin-balances in this decision as well
    if(problem.getCosts()[0] <= problem.getCosts()[1])
    {
	problem.setBestSolnNum(0);
	_splitRow = origSplitRow;
    }
    else
	problem.setBestSolnNum(1);
	    
    return true;
}

bool SplitSmallBlockHorizontally::callFMPart
		(PartitioningProblemForCapo& problem)
{
    if(_params.doRepartitioning)
	problem.setTolerance(10.0);

	//otherwise, the tolerance was already set correctly and
	//will not change

    if(_params.verb.forActions > 4)
    {
	cout<<" Targets: "<<problem.getCapacities()[0][0]<<"/"
		          <<problem.getCapacities()[1][0]<<endl;
	cout<<" Max    : "<<problem.getMaxCapacities()[0][0]<<"/"
		          <<problem.getMaxCapacities()[1][0]<<endl;
	cout<<" Min    : "<<problem.getMinCapacities()[0][0]<<"/"
		          <<problem.getMinCapacities()[1][0]<<endl;
	cout<<endl;
    }

    FMPartitioner::Parameters fmParams;
    fmParams.useEarlyStop = false;
    fmParams.useClip = (_block.getNumCells() > 75) ? 1 : 0;
    fmParams.verb.forActions  /= 4;
    fmParams.verb.forMajStats /= 4;
    fmParams.verb.forSysRes   /= 4;

    unsigned numFMStarts = 4;
    problem.reserveBuffers(numFMStarts);
    problem.getSolnBuffers().setBeginUsedSoln(0);
    problem.getSolnBuffers().setEndUsedSoln(numFMStarts);
   

    FMPartitioner partitioner1(problem, fmParams);

    double origCut = 
	problem.getCosts()[problem.getBestSolnNum()];

    if(_params.doRepartitioning)
    {
	
      if(_params.useWSTolMethod)
          problem.setPaperMethodTolerance(_splitRow);
      else
          problem.setTolerance(_params.constantTolerance);

      if(_params.verb.forActions > 4)
      {
	cout<<" Second (smaller tol) FMPart"<<endl;
	cout<<" Targets: "<<problem.getCapacities()[0][0]<<"/"
		          <<problem.getCapacities()[1][0]<<endl;
	cout<<" Max    : "<<problem.getMaxCapacities()[0][0]<<"/"
		          <<problem.getMaxCapacities()[1][0]<<endl;
	cout<<" Min    : "<<problem.getMinCapacities()[0][0]<<"/"
		          <<problem.getMinCapacities()[1][0]<<endl;
	cout<<endl;
      }

      fmParams.useClip = false;
      FMPartitioner partitioner2(problem, fmParams);

      double newCut = 
	problem.getCosts()[problem.getBestSolnNum()];

      double cutDelta = newCut - origCut;
      if(cutDelta > (origCut/10) && cutDelta >= 5)
      {
	if(_params.verb.forActions > 9)
	  cout<<" ->FM Failed.  Repartitioning increased "
	      <<"the cut too much"<<endl;
 	return false;
      }	
    }

    unsigned bestSolnNum = problem.getBestSolnNum(); 
    if(_params.verb.forMajStats > 4)
	cout<<" Best Solution had cut "<<problem.getCosts()[bestSolnNum]<<endl;
  
    if(problem.getViolations()[bestSolnNum])
    {
	if(_params.verb.forActions > 9)
	  cout<<" ->FM Failed.  No legal solutions"<<endl;
 	return false;
    }

    return true;
}

